@@ -1,17 +1,21 @@
 /* R interface to a subset of the GDAL C API for raster
    https://gdal.org/api/raster_c_api.html
-   Chris Toney <chris.toney at usda.gov> */
 
-#ifndef gdalraster_H
-#define gdalraster_H
+   Chris Toney <chris.toney at usda.gov>
+   Copyright (c) 2023-2024 gdalraster authors
+*/
 
-#include "rcpp_util.h"
+#ifndef SRC_GDALRASTER_H_
+#define SRC_GDALRASTER_H_
 
+#include <limits>
 #include <map>
 #include <string>
 #include <vector>
 
-#ifndef gdalraster_types_H
+#include "rcpp_util.h"
+
+#ifndef SRC_GDALRASTER_TYPES_H_
 #include "cpl_port.h"
 int CPL_DLL CPL_STDCALL GDALTermProgressR(double, const char *, void *);
 #endif
@@ -23,6 +27,20 @@ typedef void *GDALRasterBandH;
 typedef enum {GA_ReadOnly = 0, GA_Update = 1} GDALAccess;
 #endif
 
+// The function ARE_REAL_EQUAL was copied from gcore/gdal_priv.h since we
+// do not need that header otherwise
+// Copyright (c) 1998, Frank Warmerdam
+// Copyright (c) 2007-2014, Even Rouault <even dot rouault at spatialys.com>
+// License: MIT
+// Behavior is undefined if fVal1 or fVal2 are NaN (should be tested before
+// calling this function)
+template <class T> inline bool ARE_REAL_EQUAL(T fVal1, T fVal2, int ulp = 2)
+{
+    return fVal1 == fVal2 || /* Should cover infinity */
+           std::abs(fVal1 - fVal2) < std::numeric_limits<float>::epsilon() *
+                                         std::abs(fVal1 + fVal2) * ulp;
+}
+
 #ifdef GDAL_H_INCLUDED
 // Map certain GDAL enums to string names for use in R
 // GDALColorInterp (GCI)
@@ -68,17 +86,145 @@ const std::map<std::string, GDALRATFieldUsage> MAP_GFU{
 };
 #endif
 
+class GDALRaster {
+ private:
+    std::string fname_in;
+    Rcpp::CharacterVector open_options_in;
+    bool shared_in;
+    GDALDatasetH  hDataset;
+    GDALAccess eAccess;
+
+ public:
+    GDALRaster();
+    explicit GDALRaster(Rcpp::CharacterVector filename);
+    GDALRaster(Rcpp::CharacterVector filename, bool read_only);
+    GDALRaster(Rcpp::CharacterVector filename, bool read_only,
+               Rcpp::CharacterVector open_options);
+    GDALRaster(Rcpp::CharacterVector filename, bool read_only,
+               Rcpp::Nullable<Rcpp::CharacterVector> open_options,
+               bool shared);
+
+    // read/write fields exposed to R
+    Rcpp::CharacterVector infoOptions = Rcpp::CharacterVector::create();
+    bool quiet = false;
+    bool readByteAsRaw = false;
+
+    // methods exported to R
+    std::string getFilename() const;
+    void setFilename(std::string filename);
+    void open(bool read_only);
+    bool isOpen() const;
+    Rcpp::CharacterVector getFileList() const;
+
+    void info() const;
+    std::string infoAsJSON() const;
+
+    std::string getDriverShortName() const;
+    std::string getDriverLongName() const;
+
+    int getRasterXSize() const;
+    int getRasterYSize() const;
+    std::vector<double> getGeoTransform() const;
+    bool setGeoTransform(std::vector<double> transform);
+    int getRasterCount() const;
+
+    std::string getProjection() const;
+    std::string getProjectionRef() const;
+    bool setProjection(std::string projection);
+
+    std::vector<double> bbox() const;
+    std::vector<double> res() const;
+    std::vector<int> dim() const;
+    Rcpp::IntegerMatrix get_pixel_line(const Rcpp::RObject& xy) const;
+
+    std::vector<int> getBlockSize(int band) const;
+    std::vector<int> getActualBlockSize(int band, int xblockoff,
+                                        int yblockoff) const;
+    int getOverviewCount(int band) const;
+    void buildOverviews(std::string resampling, std::vector<int> levels,
+                        std::vector<int> bands);
+    std::string getDataTypeName(int band) const;
+    bool hasNoDataValue(int band) const;
+    double getNoDataValue(int band) const;
+    bool setNoDataValue(int band, double nodata_value);
+    void deleteNoDataValue(int band);
+    std::string getUnitType(int band) const;
+    bool setUnitType(int band, std::string unit_type);
+    bool hasScale(int band) const;
+    double getScale(int band) const;
+    bool setScale(int band, double scale);
+    bool hasOffset(int band) const;
+    double getOffset(int band) const;
+    bool setOffset(int band, double offset);
+    std::string getDescription(int band) const;
+    void setDescription(int band, std::string desc);
+    std::string getRasterColorInterp(int band) const;
+    void setRasterColorInterp(int band, std::string col_interp);
+
+    std::vector<double> getMinMax(int band, bool approx_ok) const;
+    Rcpp::NumericVector getStatistics(int band, bool approx_ok,
+                                      bool force) const;
+    void clearStatistics();
+    std::vector<double> getHistogram(int band, double min, double max,
+                                     int num_buckets, bool incl_out_of_range,
+                                     bool approx_ok) const;
+    Rcpp::List getDefaultHistogram(int band, bool force) const;
+
+    Rcpp::CharacterVector getMetadata(int band, std::string domain) const;
+    std::string getMetadataItem(int band, std::string mdi_name,
+                                std::string domain) const;
+    void setMetadataItem(int band, std::string mdi_name, std::string mdi_value,
+                         std::string domain);
+    Rcpp::CharacterVector getMetadataDomainList(int band) const;
+
+    SEXP read(int band, int xoff, int yoff, int xsize, int ysize,
+              int out_xsize, int out_ysize) const;
+
+    void write(int band, int xoff, int yoff, int xsize, int ysize,
+               const Rcpp::RObject& rasterData);
+
+    void fillRaster(int band, double value, double ivalue);
+
+    SEXP getColorTable(int band) const;
+    std::string getPaletteInterp(int band) const;
+    bool setColorTable(int band, const Rcpp::RObject& col_tbl,
+                       std::string palette_interp);
+
+    SEXP getDefaultRAT(int band) const;
+    bool setDefaultRAT(int band, const Rcpp::DataFrame& df);
+
+    void flushCache();
+
+    int getChecksum(int band, int xoff, int yoff, int xsize, int ysize) const;
+
+    void close();
+
+    // methods for internal use not exported to R
+    void _checkAccess(GDALAccess access_needed) const;
+    GDALRasterBandH _getBand(int band) const;
+    bool _readableAsInt(int band) const;
+    bool _hasInt64() const;
+    void _warnInt64() const;
+    GDALDatasetH _getGDALDatasetH() const;
+};
+
+RCPP_EXPOSED_CLASS(GDALRaster)
+
 Rcpp::CharacterVector gdal_version();
 int _gdal_version_num();
 Rcpp::DataFrame gdal_formats(std::string fmt);
 std::string get_config_option(std::string key);
 void set_config_option(std::string key, std::string value);
 int get_cache_used();
+int _dump_open_datasets(std::string outfile);
+int get_num_cpus();
+Rcpp::NumericVector get_usable_physical_ram();
 void push_error_handler(std::string handler);
 void pop_error_handler();
+bool has_spatialite();
+bool http_enabled();
 
 Rcpp::CharacterVector _check_gdal_filename(Rcpp::CharacterVector filename);
-int _get_physical_RAM();
 
 bool create(std::string format, Rcpp::CharacterVector dst_filename,
             int xsize, int ysize, int nbands, std::string dataType,
@@ -112,8 +258,8 @@ bool vsi_sync(Rcpp::CharacterVector src,
               Rcpp::CharacterVector target,
               bool show_progess,
               Rcpp::Nullable<Rcpp::CharacterVector> options);
-int vsi_mkdir(Rcpp::CharacterVector path, int mode);
-int vsi_rmdir(Rcpp::CharacterVector path);
+int vsi_mkdir(Rcpp::CharacterVector path, std::string mode, bool recursive);
+int vsi_rmdir(Rcpp::CharacterVector path, bool recursive);
 int vsi_unlink(Rcpp::CharacterVector filename);
 SEXP vsi_unlink_batch(Rcpp::CharacterVector filenames);
 SEXP vsi_stat(Rcpp::CharacterVector filename, std::string info);
@@ -124,13 +270,17 @@ bool vsi_supports_seq_write(Rcpp::CharacterVector filename,
                             bool allow_local_tmpfile);
 bool vsi_supports_rnd_write(Rcpp::CharacterVector filename,
                             bool allow_local_tmpfile);
-double vsi_get_disk_free_space(Rcpp::CharacterVector path);
+Rcpp::NumericVector vsi_get_disk_free_space(Rcpp::CharacterVector path);
+SEXP vsi_get_file_metadata(Rcpp::CharacterVector filename, std::string domain);
 
 Rcpp::NumericVector _apply_geotransform(const std::vector<double> gt,
                                         double pixel, double line);
 Rcpp::NumericVector inv_geotransform(const std::vector<double> gt);
-Rcpp::IntegerMatrix get_pixel_line(const Rcpp::NumericMatrix xy,
-                                   const std::vector<double> gt);
+Rcpp::IntegerMatrix _get_pixel_line_gt(const Rcpp::RObject& xy,
+                                       const std::vector<double> gt);
+
+Rcpp::IntegerMatrix _get_pixel_line_ds(const Rcpp::RObject& xy,
+                                       const GDALRaster* ds);
 
 bool buildVRT(Rcpp::CharacterVector vrt_filename,
               Rcpp::CharacterVector input_rasters,
@@ -146,7 +296,7 @@ Rcpp::DataFrame _combine(Rcpp::CharacterVector src_files,
                          Rcpp::Nullable<Rcpp::CharacterVector> options,
                          bool quiet);
 
-Rcpp::DataFrame _value_count(std::string src_filename, int band,
+Rcpp::DataFrame _value_count(const GDALRaster& src_ds, int band,
                              bool quiet);
 
 bool _dem_proc(std::string mode,
@@ -168,7 +318,8 @@ bool footprint(Rcpp::CharacterVector src_filename,
 bool ogr2ogr(Rcpp::CharacterVector src_dsn,
              Rcpp::CharacterVector dst_dsn,
              Rcpp::Nullable<Rcpp::CharacterVector> src_layers,
-             Rcpp::Nullable<Rcpp::CharacterVector> cl_arg);
+             Rcpp::Nullable<Rcpp::CharacterVector> cl_arg,
+             Rcpp::Nullable<Rcpp::CharacterVector> open_options);
 
 std::string ogrinfo(Rcpp::CharacterVector dsn,
                     Rcpp::Nullable<Rcpp::CharacterVector> layers,
@@ -209,113 +360,4 @@ Rcpp::IntegerMatrix createColorRamp(int start_index,
                                     Rcpp::IntegerVector end_color,
                                     std::string palette_interp);
 
-class GDALRaster {
-
-    private:
-    std::string fname_in;
-    Rcpp::CharacterVector open_options_in;
-    GDALDatasetH  hDataset;
-    GDALAccess eAccess;
-
-    public:
-    GDALRaster();
-    GDALRaster(Rcpp::CharacterVector filename);
-    GDALRaster(Rcpp::CharacterVector filename, bool read_only);
-    GDALRaster(Rcpp::CharacterVector filename, bool read_only,
-               Rcpp::CharacterVector open_options);
-
-    std::string getFilename() const;
-    void open(bool read_only);
-    bool isOpen() const;
-    Rcpp::CharacterVector getFileList() const;
-
-    void info() const;
-    std::string infoAsJSON() const;
-
-    std::string getDriverShortName() const;
-    std::string getDriverLongName() const;
-
-    int getRasterXSize() const;
-    int getRasterYSize() const;
-    std::vector<double> getGeoTransform() const;
-    bool setGeoTransform(std::vector<double> transform);
-    int getRasterCount() const;
-
-    std::string getProjectionRef() const;
-    bool setProjection(std::string projection);
-
-    std::vector<double> bbox() const;
-    std::vector<double> res() const;
-    std::vector<int> dim() const;
-
-    std::vector<int> getBlockSize(int band) const;
-    int getOverviewCount(int band) const;
-    void buildOverviews(std::string resampling, std::vector<int> levels,
-                        std::vector<int> bands);
-    std::string getDataTypeName(int band) const;
-    bool hasNoDataValue(int band) const;
-    double getNoDataValue(int band) const;
-    bool setNoDataValue(int band, double nodata_value);
-    void deleteNoDataValue(int band);
-    std::string getUnitType(int band) const;
-    bool setUnitType(int band, std::string unit_type);
-    bool hasScale(int band) const;
-    double getScale(int band) const;
-    bool setScale(int band, double scale);
-    bool hasOffset(int band) const;
-    double getOffset(int band) const;
-    bool setOffset(int band, double offset);
-    std::string getDescription(int band) const;
-    void setDescription(int band, std::string desc);
-    std::string getRasterColorInterp(int band) const;
-    void setRasterColorInterp(int band, std::string col_interp);
-
-    std::vector<double> getMinMax(int band, bool approx_ok) const;
-    Rcpp::NumericVector getStatistics(int band,	bool approx_ok,
-                                      bool force) const;
-    void clearStatistics();
-    std::vector<double> getHistogram(int band, double min, double max,
-                                     int num_buckets, bool incl_out_of_range,
-                                     bool approx_ok) const;
-    Rcpp::List getDefaultHistogram(int band, bool force) const;
-
-    Rcpp::CharacterVector getMetadata(int band, std::string domain) const;
-    std::string getMetadataItem(int band, std::string mdi_name,
-                                std::string domain) const;
-    void setMetadataItem(int band, std::string mdi_name, std::string mdi_value,
-                         std::string domain);
-    Rcpp::CharacterVector getMetadataDomainList(int band) const;
-
-    SEXP read(int band, int xoff, int yoff, int xsize, int ysize,
-              int out_xsize, int out_ysize) const;
-
-    void write(int band, int xoff, int yoff, int xsize, int ysize,
-               Rcpp::RObject rasterData);
-
-    void fillRaster(int band, double value, double ivalue);
-
-    SEXP getColorTable(int band) const;
-    std::string getPaletteInterp(int band) const;
-    bool setColorTable(int band, Rcpp::RObject& col_tbl,
-                       std::string palette_interp);
-
-    SEXP getDefaultRAT(int band) const;
-    bool setDefaultRAT(int band, Rcpp::DataFrame& df);
-
-    void flushCache();
-
-    int getChecksum(int band, int xoff, int yoff, int xsize, int ysize) const;
-
-    void close();
-
-    // methods for internal use not exported to R
-    void _checkAccess(GDALAccess access_needed) const;
-    GDALRasterBandH _getBand(int band) const;
-    bool _readableAsInt(int band) const;
-    bool _hasInt64() const;
-    void _warnInt64() const;
-};
-
-RCPP_EXPOSED_CLASS(GDALRaster)
-
-#endif
+#endif  // SRC_GDALRASTER_H_
