@@ -4,8 +4,9 @@
 
 #include "cpl_conv.h"
 #include "ogr_api.h"
-#include "ogr_srs_api.h"
+#include "ogr_core.h"
 #include "ogr_spatialref.h"
+#include "ogr_srs_api.h"
 
 #include "geos_wkt.h"
 
@@ -39,14 +40,16 @@ bool has_geos() {
 // test if GDAL built against GEOS
 // this is now obsolete:
 // GDAL built against GEOS is required at gdalraster 1.10
-    OGRGeometryH hGeom = OGR_G_CreateGeometry(wkbPoint);
+    OGRGeometryH hGeom = nullptr;
+    hGeom = OGR_G_CreateGeometry(wkbPoint);
     if (hGeom == nullptr)
         Rcpp::stop("failed to create geometry object");
     OGR_G_SetPoint_2D(hGeom, 0, 0, 0);
 
     // If GDAL is built without the GEOS library, this function will
     // always return FALSE:
-    bool ret = OGR_G_IsSimple(hGeom);
+    bool ret = false;
+    ret = OGR_G_IsSimple(hGeom);
     OGR_G_DestroyGeometry(hGeom);
     return ret;
 }
@@ -59,25 +62,33 @@ bool has_geos() {
 // [[Rcpp::export(name = ".g_create")]]
 std::string _g_create(Rcpp::NumericMatrix xy, std::string geom_type) {
 // Create a geometry from a list of points (vertices).
-// Currently only for POINT, LINESTRING, POLYGON.
-// Only simple polygons composed of one exterior ring are supported.
+// Currently for POINT, MULTIPOINT, LINESTRING, POLYGON.
+// Only simple polygons composed of one ring are supported.
 
     OGRGeometryH hGeom = nullptr;
+    OGRGeometryH hPoly = nullptr;
+    OGRErr err = OGRERR_NONE;
+    char *pszWKT = nullptr;
+    std::string wkt = "";
 
-    if (geom_type == "point" || geom_type == "POINT") {
+    if (EQUALN(geom_type.c_str(), "POINT", 5)) {
         geom_type = "POINT";
         hGeom = OGR_G_CreateGeometry(wkbPoint);
     }
-    else if (geom_type == "linestring" || geom_type == "LINESTRING") {
+    else if (EQUALN(geom_type.c_str(), "MULTIPOINT", 10)) {
+        geom_type = "MULTIPOINT";
+        hGeom = OGR_G_CreateGeometry(wkbMultiPoint);
+    }
+    else if (EQUALN(geom_type.c_str(), "LINESTRING", 10)) {
         geom_type = "LINESTRING";
         hGeom = OGR_G_CreateGeometry(wkbLineString);
     }
-    else if (geom_type == "polygon" || geom_type == "POLYGON") {
+    else if (EQUALN(geom_type.c_str(), "POLYGON", 7)) {
         geom_type = "POLYGON";
         hGeom = OGR_G_CreateGeometry(wkbLinearRing);
     }
     else {
-        Rcpp::stop("geometry type is not valid");
+        Rcpp::stop("geometry type not supported");
     }
 
     if (hGeom == nullptr)
@@ -100,37 +111,124 @@ std::string _g_create(Rcpp::NumericMatrix xy, std::string geom_type) {
         }
         if (geom_type == "POLYGON" && nPts < 4) {
             OGR_G_DestroyGeometry(hGeom);
-            Rcpp::stop("polygon geometry must have at least four points");
+            Rcpp::stop("polygon must have at least four points");
         }
 
-        OGR_G_SetPointCount(hGeom, (int) nPts);
-        for (R_xlen_t i=0; i < nPts; ++i)
-            OGR_G_SetPoint_2D(hGeom, i, xy(i, 0), xy(i, 1));
+        if (geom_type == "MULTIPOINT") {
+            for (R_xlen_t i=0; i < nPts; ++i) {
+                OGRGeometryH hPt = OGR_G_CreateGeometry(wkbPoint);
+                OGR_G_SetPoint_2D(hPt, 0, xy(i, 0), xy(i, 1));
+                err = OGR_G_AddGeometryDirectly(hGeom, hPt);
+                if (err != OGRERR_NONE) {
+                    if (hGeom != nullptr)
+                        OGR_G_DestroyGeometry(hGeom);
+                    Rcpp::stop("failed to add POINT to MULTIPOINT");
+                }
+            }
+        }
+        else {
+            OGR_G_SetPointCount(hGeom, static_cast<int>(nPts));
+            for (R_xlen_t i=0; i < nPts; ++i) {
+                OGR_G_SetPoint_2D(hGeom, i, xy(i, 0), xy(i, 1));
+            }
+        }
     }
 
     if (geom_type == "POLYGON") {
-        OGRGeometryH hPoly = OGR_G_CreateGeometry(wkbPolygon);
-
+        hPoly = OGR_G_CreateGeometry(wkbPolygon);
+        if (hPoly == nullptr) {
+            if (hGeom != nullptr)
+                OGR_G_DestroyGeometry(hGeom);
+            Rcpp::stop("failed to create polygon geometry object");
+        }
         CPLSetConfigOption("OGR_GEOMETRY_ACCEPT_UNCLOSED_RING", "NO");
-        if (OGR_G_AddGeometryDirectly(hPoly, hGeom) != OGRERR_NONE)
-            Rcpp::stop("failed to create polygon geometry (unclosed ring?)");
+        err = OGR_G_AddGeometryDirectly(hPoly, hGeom);
         CPLSetConfigOption("OGR_GEOMETRY_ACCEPT_UNCLOSED_RING", nullptr);
+        if (err != OGRERR_NONE) {
+            if (hPoly != nullptr)
+                OGR_G_DestroyGeometry(hPoly);
+            Rcpp::stop("failed to create polygon geometry (unclosed ring?)");
+        }
 
-        char* pszWKT;
         OGR_G_ExportToWkt(hPoly, &pszWKT);
-        std::string wkt(pszWKT);
-        CPLFree(pszWKT);
         OGR_G_DestroyGeometry(hPoly);
-        return wkt;
     }
     else {
-        char* pszWKT;
         OGR_G_ExportToWkt(hGeom, &pszWKT);
-        std::string wkt(pszWKT);
-        CPLFree(pszWKT);
         OGR_G_DestroyGeometry(hGeom);
-        return wkt;
     }
+
+    if (pszWKT != nullptr) {
+        wkt = pszWKT;
+        CPLFree(pszWKT);
+    }
+    return wkt;
+}
+
+//' @noRd
+// [[Rcpp::export(name = ".g_add_geom")]]
+std::string _g_add_geom(std::string sub_geom, std::string container) {
+// Add a geometry to a geometry container.
+// LINEARRING (as POLYGON) to POLYGON, POINT to MULTIPOINT, LINESTRING to
+// MULTILINESTRING, or POLYGON to MULTIPOLYGON
+
+    OGRGeometryH hSubGeom = nullptr;
+    OGRGeometryH hGeom = nullptr;
+    OGRErr err = OGRERR_NONE;
+    char* pszWKT_sub = (char*) sub_geom.c_str();
+    char* pszWKT_container = (char*) container.c_str();
+
+    err = OGR_G_CreateFromWkt(&pszWKT_sub, nullptr, &hSubGeom);
+    if (err != OGRERR_NONE || hSubGeom == nullptr) {
+        if (hSubGeom != nullptr)
+            OGR_G_DestroyGeometry(hSubGeom);
+        Rcpp::stop("failed to create geometry object for 'sub_geom'");
+    }
+
+    err = OGR_G_CreateFromWkt(&pszWKT_container, nullptr, &hGeom);
+    if (err != OGRERR_NONE || hGeom == nullptr) {
+        if (hGeom != nullptr)
+            OGR_G_DestroyGeometry(hGeom);
+        OGR_G_DestroyGeometry(hSubGeom);
+        Rcpp::stop("failed to create geometry object for 'container'");
+    }
+
+    CPLSetConfigOption("OGR_GEOMETRY_ACCEPT_UNCLOSED_RING", "NO");
+
+    if (EQUALN(OGR_G_GetGeometryName(hGeom), "POLYGON", 7) &&
+            EQUALN(OGR_G_GetGeometryName(hSubGeom), "POLYGON", 7)) {
+        // interpret sub_geom as one linearring
+
+        OGRGeometryH hRing = OGR_G_GetGeometryRef(hSubGeom, 0);
+        err = OGR_G_AddGeometry(hGeom, hRing);
+        if (err != OGRERR_NONE) {
+            if (hGeom != nullptr)
+                OGR_G_DestroyGeometry(hGeom);
+            if (hSubGeom != nullptr)
+                OGR_G_DestroyGeometry(hSubGeom);
+            Rcpp::stop("failed to add 'sub_geom' to 'container'");
+        }
+    }
+    else {
+        err = OGR_G_AddGeometryDirectly(hGeom, hSubGeom);
+        if (err != OGRERR_NONE) {
+            if (hGeom != nullptr)
+                OGR_G_DestroyGeometry(hGeom);
+            Rcpp::stop("failed to add 'sub_geom' to 'container'");
+        }
+    }
+
+    CPLSetConfigOption("OGR_GEOMETRY_ACCEPT_UNCLOSED_RING", nullptr);
+
+    char* pszWKT = nullptr;
+    OGR_G_ExportToWkt(hGeom, &pszWKT);
+    std::string wkt = "";
+    if (pszWKT != nullptr) {
+        wkt = pszWKT;
+        CPLFree(pszWKT);
+    }
+    OGR_G_DestroyGeometry(hGeom);
+    return wkt;
 }
 
 //' @noRd
@@ -142,14 +240,62 @@ bool _g_is_valid(std::string geom) {
 // this function will always return FALSE.
 
     OGRGeometryH hGeom = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT = (char*) geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT, nullptr, &hGeom);
-    if (err != OGRERR_NONE || hGeom == nullptr)
+    if (err != OGRERR_NONE || hGeom == nullptr) {
+        if (hGeom != nullptr)
+            OGR_G_DestroyGeometry(hGeom);
         Rcpp::stop("failed to create geometry object from WKT string");
+    }
 
-    bool ret = OGR_G_IsValid(hGeom);
+    bool ret = false;
+    ret = OGR_G_IsValid(hGeom);
+    OGR_G_DestroyGeometry(hGeom);
+    return ret;
+}
+
+//' @noRd
+// [[Rcpp::export(name = ".g_is_empty")]]
+bool _g_is_empty(std::string geom) {
+// Test if the geometry is empty.
+
+    OGRGeometryH hGeom = nullptr;
+    OGRErr err = OGRERR_NONE;
+    char* pszWKT = (char*) geom.c_str();
+
+    err = OGR_G_CreateFromWkt(&pszWKT, nullptr, &hGeom);
+    if (err != OGRERR_NONE || hGeom == nullptr) {
+        if (hGeom != nullptr)
+            OGR_G_DestroyGeometry(hGeom);
+        Rcpp::stop("failed to create geometry object from WKT string");
+    }
+
+    bool ret = false;
+    ret = OGR_G_IsEmpty(hGeom);
+    OGR_G_DestroyGeometry(hGeom);
+    return ret;
+}
+
+//' @noRd
+// [[Rcpp::export(name = ".g_name")]]
+std::string _g_name(std::string geom) {
+// extract the geometry type name from a WKT geometry
+
+    OGRGeometryH hGeom = nullptr;
+    OGRErr err = OGRERR_NONE;
+    char* pszWKT = (char*) geom.c_str();
+
+    err = OGR_G_CreateFromWkt(&pszWKT, nullptr, &hGeom);
+    if (err != OGRERR_NONE || hGeom == nullptr) {
+        if (hGeom != nullptr)
+            OGR_G_DestroyGeometry(hGeom);
+        Rcpp::stop("failed to create geometry object from WKT string");
+    }
+
+    std::string ret = "";
+    ret = OGR_G_GetGeometryName(hGeom);
     OGR_G_DestroyGeometry(hGeom);
     return ret;
 }
@@ -167,17 +313,23 @@ bool _g_intersects(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
@@ -200,17 +352,23 @@ bool _g_equals(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
@@ -233,17 +391,23 @@ bool _g_disjoint(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
@@ -266,17 +430,23 @@ bool _g_touches(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
@@ -299,17 +469,23 @@ bool _g_contains(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
@@ -332,17 +508,23 @@ bool _g_within(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
@@ -365,17 +547,23 @@ bool _g_crosses(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
@@ -389,7 +577,8 @@ bool _g_crosses(std::string this_geom, std::string other_geom) {
 // [[Rcpp::export(name = ".g_overlaps")]]
 bool _g_overlaps(std::string this_geom, std::string other_geom) {
 // Tests if this geometry and the other geometry overlap, that is their
-// intersection has a non-zero area.
+// intersection has a non-zero area (they have some but not all points in
+// common).
 // Geometry validity is not checked. In case you are unsure of the validity
 // of the input geometries, call _g_is_valid() before, otherwise the result
 // might be wrong.
@@ -399,17 +588,23 @@ bool _g_overlaps(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
@@ -426,7 +621,7 @@ bool _g_overlaps(std::string this_geom, std::string other_geom) {
 //' @noRd
 // [[Rcpp::export(name = ".g_buffer")]]
 std::string _g_buffer(std::string geom, double dist, int quad_segs = 30) {
-//Compute buffer of geometry.
+// Compute buffer of geometry.
 
 // Builds a new geometry containing the buffer region around the geometry on
 // which it is invoked. The buffer is a polygon containing the region within
@@ -440,23 +635,30 @@ std::string _g_buffer(std::string geom, double dist, int quad_segs = 30) {
 // reduce the accuracy of the result.
 
     OGRGeometryH hGeom = nullptr;
-    OGRErr err;
+    OGRGeometryH hBufferGeom = nullptr;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT = (char*) geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT, nullptr, &hGeom);
-    if (err != OGRERR_NONE || hGeom == nullptr)
+    if (err != OGRERR_NONE || hGeom == nullptr) {
+        if (hGeom != nullptr)
+            OGR_G_DestroyGeometry(hGeom);
         Rcpp::stop("failed to create geometry object from WKT string");
+    }
 
-    OGRGeometryH hBufferGeom = OGR_G_Buffer(hGeom, dist, quad_segs);
+    hBufferGeom = OGR_G_Buffer(hGeom, dist, quad_segs);
     if (hBufferGeom == nullptr) {
         OGR_G_DestroyGeometry(hGeom);
         Rcpp::stop("failed to create buffer geometry");
     }
 
-    char* pszWKT_out;
+    char *pszWKT_out = nullptr;
     OGR_G_ExportToWkt(hBufferGeom, &pszWKT_out);
-    std::string wkt_out(pszWKT_out);
-    CPLFree(pszWKT_out);
+    std::string wkt_out = "";
+    if (pszWKT_out != nullptr) {
+        wkt_out = pszWKT_out;
+        CPLFree(pszWKT_out);
+    }
     OGR_G_DestroyGeometry(hGeom);
     OGR_G_DestroyGeometry(hBufferGeom);
 
@@ -482,31 +684,41 @@ std::string _g_intersection(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
-    OGRGeometryH hGeom = OGR_G_Intersection(hGeom_this, hGeom_other);
+    OGRGeometryH hGeom = nullptr;
+    hGeom = OGR_G_Intersection(hGeom_this, hGeom_other);
     if (hGeom == nullptr) {
         OGR_G_DestroyGeometry(hGeom_this);
         OGR_G_DestroyGeometry(hGeom_other);
         return "";
     }
 
-    char* pszWKT_out;
+    char* pszWKT_out = nullptr;
     OGR_G_ExportToWkt(hGeom, &pszWKT_out);
-    std::string wkt_out(pszWKT_out);
-    CPLFree(pszWKT_out);
+    std::string wkt_out = "";
+    if (pszWKT_out != nullptr) {
+        wkt_out = pszWKT_out;
+        CPLFree(pszWKT_out);
+    }
     OGR_G_DestroyGeometry(hGeom);
     OGR_G_DestroyGeometry(hGeom_this);
     OGR_G_DestroyGeometry(hGeom_other);
@@ -528,31 +740,41 @@ std::string _g_union(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
-    OGRGeometryH hGeom = OGR_G_Union(hGeom_this, hGeom_other);
+    OGRGeometryH hGeom = nullptr;
+    hGeom = OGR_G_Union(hGeom_this, hGeom_other);
     if (hGeom == nullptr) {
         OGR_G_DestroyGeometry(hGeom_this);
         OGR_G_DestroyGeometry(hGeom_other);
         return "";
     }
 
-    char* pszWKT_out;
+    char* pszWKT_out = nullptr;
     OGR_G_ExportToWkt(hGeom, &pszWKT_out);
-    std::string wkt_out(pszWKT_out);
-    CPLFree(pszWKT_out);
+    std::string wkt_out = "";
+    if (pszWKT_out != nullptr) {
+        wkt_out = pszWKT_out;
+        CPLFree(pszWKT_out);
+    }
     OGR_G_DestroyGeometry(hGeom);
     OGR_G_DestroyGeometry(hGeom_this);
     OGR_G_DestroyGeometry(hGeom_other);
@@ -574,31 +796,41 @@ std::string _g_difference(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
-    OGRGeometryH hGeom = OGR_G_Difference(hGeom_this, hGeom_other);
+    OGRGeometryH hGeom = nullptr;
+    hGeom = OGR_G_Difference(hGeom_this, hGeom_other);
     if (hGeom == nullptr) {
         OGR_G_DestroyGeometry(hGeom_this);
         OGR_G_DestroyGeometry(hGeom_other);
         return "";
     }
 
-    char* pszWKT_out;
+    char* pszWKT_out = nullptr;
     OGR_G_ExportToWkt(hGeom, &pszWKT_out);
-    std::string wkt_out(pszWKT_out);
-    CPLFree(pszWKT_out);
+    std::string wkt_out = "";
+    if (pszWKT_out != nullptr) {
+        wkt_out = pszWKT_out;
+        CPLFree(pszWKT_out);
+    }
     OGR_G_DestroyGeometry(hGeom);
     OGR_G_DestroyGeometry(hGeom_this);
     OGR_G_DestroyGeometry(hGeom_other);
@@ -620,31 +852,41 @@ std::string _g_sym_difference(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
-    OGRGeometryH hGeom = OGR_G_SymDifference(hGeom_this, hGeom_other);
+    OGRGeometryH hGeom = nullptr;
+    hGeom = OGR_G_SymDifference(hGeom_this, hGeom_other);
     if (hGeom == nullptr) {
         OGR_G_DestroyGeometry(hGeom_this);
         OGR_G_DestroyGeometry(hGeom_other);
         return "";
     }
 
-    char* pszWKT_out;
+    char* pszWKT_out = nullptr;
     OGR_G_ExportToWkt(hGeom, &pszWKT_out);
-    std::string wkt_out(pszWKT_out);
-    CPLFree(pszWKT_out);
+    std::string wkt_out = "";
+    if (pszWKT_out != nullptr) {
+        wkt_out = pszWKT_out;
+        CPLFree(pszWKT_out);
+    }
     OGR_G_DestroyGeometry(hGeom);
     OGR_G_DestroyGeometry(hGeom_this);
     OGR_G_DestroyGeometry(hGeom_other);
@@ -668,21 +910,28 @@ double _g_distance(std::string this_geom, std::string other_geom) {
 
     OGRGeometryH hGeom_this = nullptr;
     OGRGeometryH hGeom_other = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT_this = (char*) this_geom.c_str();
     char* pszWKT_other = (char*) other_geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT_this, nullptr, &hGeom_this);
-    if (err != OGRERR_NONE || hGeom_this == nullptr)
+    if (err != OGRERR_NONE || hGeom_this == nullptr) {
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
         Rcpp::stop("failed to create geometry object from first WKT string");
+    }
 
     err = OGR_G_CreateFromWkt(&pszWKT_other, nullptr, &hGeom_other);
     if (err != OGRERR_NONE || hGeom_other == nullptr) {
-        OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_this != nullptr)
+            OGR_G_DestroyGeometry(hGeom_this);
+        if (hGeom_other != nullptr)
+            OGR_G_DestroyGeometry(hGeom_other);
         Rcpp::stop("failed to create geometry object from second WKT string");
     }
 
-    double ret = OGR_G_Distance(hGeom_this, hGeom_other);
+    double ret = -1;
+    ret = OGR_G_Distance(hGeom_this, hGeom_other);
     OGR_G_DestroyGeometry(hGeom_this);
     OGR_G_DestroyGeometry(hGeom_other);
     return ret;
@@ -699,10 +948,14 @@ double _g_length(std::string geom) {
     char* pszWKT = (char*) geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT, nullptr, &hGeom);
-    if (err != OGRERR_NONE || hGeom == nullptr)
+    if (err != OGRERR_NONE || hGeom == nullptr) {
+        if (hGeom != nullptr)
+            OGR_G_DestroyGeometry(hGeom);
         Rcpp::stop("failed to create geometry object from WKT string");
+    }
 
-    double ret = OGR_G_Length(hGeom);
+    double ret = 0;
+    ret = OGR_G_Length(hGeom);
     OGR_G_DestroyGeometry(hGeom);
     return ret;
 }
@@ -718,10 +971,14 @@ double _g_area(std::string geom) {
     char* pszWKT = (char*) geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT, nullptr, &hGeom);
-    if (err != OGRERR_NONE || hGeom == nullptr)
+    if (err != OGRERR_NONE || hGeom == nullptr) {
+        if (hGeom != nullptr)
+            OGR_G_DestroyGeometry(hGeom);
         Rcpp::stop("failed to create geometry object from WKT string");
+    }
 
-    double ret = OGR_G_Area(hGeom);
+    double ret = 0;
+    ret = OGR_G_Area(hGeom);
     OGR_G_DestroyGeometry(hGeom);
     return ret;
 }
@@ -741,14 +998,18 @@ Rcpp::NumericVector _g_centroid(std::string geom) {
 // function will always fail, issuing a CPLE_NotSupported error.
 
     OGRGeometryH hGeom = nullptr;
-    OGRErr err;
+    OGRErr err = OGRERR_NONE;
     char* pszWKT = (char*) geom.c_str();
 
     err = OGR_G_CreateFromWkt(&pszWKT, nullptr, &hGeom);
-    if (err != OGRERR_NONE || hGeom == nullptr)
+    if (err != OGRERR_NONE || hGeom == nullptr) {
+        if (hGeom != nullptr)
+            OGR_G_DestroyGeometry(hGeom);
         Rcpp::stop("failed to create geometry object from WKT string");
+    }
 
-    OGRGeometryH hPoint = OGR_G_CreateGeometry(wkbPoint);
+    OGRGeometryH hPoint = nullptr;
+    hPoint = OGR_G_CreateGeometry(wkbPoint);
     if (hPoint == nullptr) {
         Rcpp::stop("failed to create point geometry object");
         OGR_G_DestroyGeometry(hGeom);
@@ -775,7 +1036,8 @@ Rcpp::NumericVector _g_centroid(std::string geom) {
 //' @noRd
 // [[Rcpp::export(name = ".g_transform")]]
 std::string _g_transform(std::string geom, std::string srs_from,
-        std::string srs_to) {
+        std::string srs_to, bool wrap_date_line = false,
+        int date_line_offset = 10) {
 // Returns a transformed geometry as WKT
 // Apply arbitrary coordinate transformation to geometry.
 // This function will transform the coordinates of a geometry from their
@@ -788,49 +1050,83 @@ std::string _g_transform(std::string geom, std::string srs_from,
 // OGRCoordinateTransformation object, and the actual SRS of the geometry will
 // be ignored. On successful completion the output OGRSpatialReference of the
 // OGRCoordinateTransformation will be assigned to the geometry.
-// This function only does reprojection on a point-by-point basis. It does not
-// include advanced logic to deal with discontinuities at poles or antimeridian.
-// For that, use the OGR_GeomTransformer_Create() and
-// OGR_GeomTransformer_Transform() functions.
+// This function uses the OGR_GeomTransformer_Create() and
+// OGR_GeomTransformer_Transform() functions: this is a enhanced version of
+// OGR_G_Transform(). When reprojecting geometries from a Polar Stereographic
+// projection or a projection naturally crossing the antimeridian (like UTM
+// Zone 60) to a geographic CRS, it will cut geometries along the antimeridian.
+// So a LineString might be returned as a MultiLineString.
 
     OGRSpatialReference oSourceSRS, oDestSRS;
-    OGRCoordinateTransformation *poCT;
-    OGRErr err;
-
-    err = oSourceSRS.importFromWkt(srs_from.c_str());
-    if (err != OGRERR_NONE)
-        Rcpp::stop("failed to import source SRS from WKT string");
-
-    err = oDestSRS.importFromWkt(srs_to.c_str());
-    if (err != OGRERR_NONE)
-        Rcpp::stop("failed to import destination SRS from WKT string");
-
-    poCT = OGRCreateCoordinateTransformation(&oSourceSRS, &oDestSRS);
-    if (poCT == nullptr)
-        Rcpp::stop("failed to create coordinate transformer");
-
+    OGRCoordinateTransformation *poCT = nullptr;
+    OGRGeomTransformerH hGeomTransformer = nullptr;
     OGRGeometryH hGeom = nullptr;
+    OGRGeometryH hGeom2 = nullptr;
+    OGRErr err = OGRERR_NONE;
+
     char* pszWKT = (char*) geom.c_str();
     err = OGR_G_CreateFromWkt(&pszWKT, nullptr, &hGeom);
     if (err != OGRERR_NONE || hGeom == nullptr) {
-        OGRCoordinateTransformation::DestroyCT(poCT);
+        if (hGeom != nullptr)
+            OGR_G_DestroyGeometry(hGeom);
         Rcpp::stop("failed to create geometry object from WKT string");
     }
 
-    err = OGR_G_Transform(hGeom, OGRCoordinateTransformation::ToHandle(poCT));
+    err = oSourceSRS.importFromWkt(srs_from.c_str());
     if (err != OGRERR_NONE) {
+        OGR_G_DestroyGeometry(hGeom);
+        Rcpp::stop("failed to import source SRS from WKT string");
+    }
+
+    err = oDestSRS.importFromWkt(srs_to.c_str());
+    if (err != OGRERR_NONE) {
+        OGR_G_DestroyGeometry(hGeom);
+        Rcpp::stop("failed to import destination SRS from WKT string");
+    }
+
+    poCT = OGRCreateCoordinateTransformation(&oSourceSRS, &oDestSRS);
+    if (poCT == nullptr) {
+        OGR_G_DestroyGeometry(hGeom);
+        Rcpp::stop("failed to create coordinate transformer");
+    }
+
+    std::vector<char *> options;
+    std::string offset;
+    if (wrap_date_line) {
+        options.push_back((char *) "WRAPDATELINE=YES");
+        offset = "DATELINEOFFSET=" + std::to_string(date_line_offset);
+        options.push_back((char *) offset.c_str());
+    }
+    options.push_back(nullptr);
+
+    hGeomTransformer = OGR_GeomTransformer_Create(
+            OGRCoordinateTransformation::ToHandle(poCT), options.data());
+    if (hGeomTransformer == nullptr) {
         OGRCoordinateTransformation::DestroyCT(poCT);
         OGR_G_DestroyGeometry(hGeom);
+        Rcpp::stop("failed to create geometry transformer");
+    }
+
+    hGeom2 = OGR_GeomTransformer_Transform(hGeomTransformer, hGeom);
+    if (hGeom2 == nullptr) {
+        OGRCoordinateTransformation::DestroyCT(poCT);
+        OGR_GeomTransformer_Destroy(hGeomTransformer);
+        OGR_G_DestroyGeometry(hGeom);
         Rcpp::stop("transformation failed");
     }
 
-    char* pszWKT_out;
-    OGR_G_ExportToWkt(hGeom, &pszWKT_out);
-    std::string wkt_out(pszWKT_out);
-    CPLFree(pszWKT_out);
+    char* pszWKT_out = nullptr;
+    OGR_G_ExportToWkt(hGeom2, &pszWKT_out);
+    std::string wkt_out = "";
+    if (pszWKT_out != nullptr) {
+        wkt_out = pszWKT_out;
+        CPLFree(pszWKT_out);
+    }
 
     OGRCoordinateTransformation::DestroyCT(poCT);
+    OGR_GeomTransformer_Destroy(hGeomTransformer);
     OGR_G_DestroyGeometry(hGeom);
+    OGR_G_DestroyGeometry(hGeom2);
 
     return wkt_out;
 }
