@@ -3,18 +3,22 @@
    For spatial predicate definitions: https://en.wikipedia.org/wiki/DE-9IM
    Chris Toney <chris.toney at usda.gov> */
 
-#ifndef geos_wkt_H
-#define geos_wkt_H
+#ifndef SRC_GEOS_WKT_H_
+#define SRC_GEOS_WKT_H_
+
+#include <string>
+#include <vector>
 
 #include <Rcpp.h>
 
-#include <string>
-
 std::vector<int> _getGEOSVersion();
-bool has_geos(); // GDAL built against GEOS is required at gdalraster 1.10
+bool has_geos();  // GDAL built against GEOS is required at gdalraster 1.10
 
 std::string _g_create(Rcpp::NumericMatrix xy, std::string geom_type);
+std::string _g_add_geom(std::string sub_geom, std::string container);
 bool _g_is_valid(std::string geom);
+bool _g_is_empty(std::string geom);
+std::string _g_name(std::string geom);
 
 bool _g_intersects(std::string this_geom, std::string other_geom);
 bool _g_equals(std::string this_geom, std::string other_geom);
@@ -38,7 +42,7 @@ double _g_area(std::string geom);
 Rcpp::NumericVector _g_centroid(std::string geom);
 
 std::string _g_transform(std::string geom, std::string srs_from,
-        std::string srs_to);
-
-#endif
+                         std::string srs_to, bool wrap_date_line,
+                         int date_line_offset);
 
+#endif  // SRC_GEOS_WKT_H_
