@@ -11,14 +11,14 @@ test_that("gdal_formats returns a data frame", {
     expect_true(nrow(x) > 1)
 })
 
-test_that("_check_gdal_filename works", {
+test_that(".check_gdal_filename works", {
     elev_file <- system.file("extdata/storml_elev.tif", package="gdalraster")
     b5_file <- system.file("extdata/sr_b5_20200829.tif", package="gdalraster")
     expect_error(.check_gdal_filename(c(elev_file, b5_file)))
     vsifn <- paste0("/vsi/", b5_file)
     expect_equal(vsifn, .check_gdal_filename(vsifn))
     fn <- "~/_r82jRwnT.test"
-    expect_warning(fn_out <- .check_gdal_filename(fn))
+    fn_out <- .check_gdal_filename(fn)
     expect_equal(basename(fn_out), basename(fn))
 })
 
@@ -33,8 +33,24 @@ test_that("get_cache_used returns integer", {
     expect_type(get_cache_used(), "integer")
 })
 
-test_that(".get_physical_RAM returns integer", {
-    expect_type(.get_physical_RAM(), "integer")
+test_that("get_num_cpus returns integer", {
+    expect_type(get_num_cpus(), "integer")
+})
+
+test_that("get_usable_physical_ram returns integer64", {
+    expect_s3_class(get_usable_physical_ram(), "integer64")
+})
+
+test_that("has_spatialite returns logical", {
+    expect_type(has_spatialite(), "logical")
+})
+
+test_that("http_enabled returns logical", {
+    expect_type(http_enabled(), "logical")
+})
+
+test_that(".cpl_http_cleanup runs without error", {
+    expect_no_error(.cpl_http_cleanup())
 })
 
 test_that("createCopy writes correct output", {
@@ -42,14 +58,14 @@ test_that("createCopy writes correct output", {
     tif_file <- paste0(tempdir(), "/", "storml_lndscp.tif")
     options <- c("COMPRESS=LZW")
     createCopy(format="GTiff", dst_filename=tif_file, src_filename=lcp_file,
-                options=options)
+               options=options)
     ds <- new(GDALRaster, tif_file, read_only=FALSE)
     files <- ds$getFileList()
     on.exit(unlink(files))
     md <- ds$getMetadata(band=0, domain="IMAGE_STRUCTURE")
     expect_equal(md, c("COMPRESSION=LZW", "INTERLEAVE=PIXEL"))
     for (band in 1:ds$getRasterCount())
-        ds$setNoDataValue(band, -9999)
+         ds$setNoDataValue(band, -9999)
     stats <- ds$getStatistics(band=1, approx_ok=FALSE, force=TRUE)
     expect_equal(round(stats), round(c(2438.0, 3046.0, 2675.9713, 133.0185)))
     dm <- ds$dim()
@@ -62,13 +78,19 @@ test_that("get_pixel_line gives correct results", {
     pt_file <- system.file("extdata/storml_pts.csv", package="gdalraster")
     pts <- read.csv(pt_file)
     pix_line <- c(39, 23, 1, 58, 74, 94, 68, 92, 141, 23, 57, 68, 58, 52, 90,
-                    38, 31, 85, 20, 39)
+                  38, 31, 85, 20, 39)
     raster_file <- system.file("extdata/storm_lake.lcp", package="gdalraster")
     ds <- new(GDALRaster, raster_file, read_only=TRUE)
     gt <- ds$getGeoTransform()
-    ds$close()
-    res <- get_pixel_line(as.matrix(pts[,-1]), gt)
+    res <- get_pixel_line(as.matrix(pts[, -1]), gt)
     expect_equal(as.vector(res), pix_line)
+
+    pts[11, ] <- c(11, 323318, 5105104)
+    expect_warning(res2 <- ds$get_pixel_line(pts[, -1]))
+    res <- rbind(res, c(NA, NA))
+    expect_equal(res2, res)
+
+    ds$close()
 })
 
 test_that("_apply_geotransform gives correct result", {
@@ -152,9 +174,9 @@ test_that("sieveFilter runs without error", {
 
 test_that("createColorRamp works", {
     colors <- createColorRamp(start_index = 0,
-                start_color = c(211, 211, 211),
-                end_index = 100,
-                end_color = c(0, 100, 0))
+                              start_color = c(211, 211, 211),
+                              end_index = 100,
+                              end_color = c(0, 100, 0))
     expect_equal(nrow(colors), 101)
     # non-zero start_index
     colors = createColorRamp(109, c(254, 231, 152), 127, c(254, 254, 189))
@@ -312,6 +334,14 @@ test_that("ogr2ogr works", {
     deleteDataset(ynp_wgs84)
     fld_idx <- -1
 
+    # with open_options
+    vsi_curl_clear_cache()
+    ogr2ogr(src, ynp_wgs84, cl_arg = args, open_options = "LIST_ALL_TABLES=NO")
+    fld_idx <- .ogr_field_index(ynp_wgs84, "mtbs_perims", "ig_year")
+    expect_equal(fld_idx, 8)
+    deleteDataset(ynp_wgs84)
+    fld_idx <- -1
+
     # clip to a bounding box (xmin, ymin, xmax, ymax)
     ynp_clip <- file.path(tempdir(), "ynp_fires_aoi_clip.gpkg")
     bb <- c(469685.97, 11442.45, 544069.63, 85508.15)
@@ -330,7 +360,6 @@ test_that("ogr2ogr works", {
     fld_idx <- .ogr_field_index(ynp_filtered, "mtbs_perims", "ig_year")
     expect_equal(fld_idx, 8)
     deleteDataset(ynp_filtered)
-
 })
 
 test_that("ogrinfo works", {
@@ -342,6 +371,9 @@ test_that("ogrinfo works", {
     expect_vector(info, ptype = character(), size = 1)
     expect_false(info[1] == "")
 
+    json <- ogrinfo(src, "mtbs_perims", cl_arg = c("-json", "-so", "-nomd"))
+    expect_true(nchar(json) > 1000)
+
     src_mem <- paste0("/vsimem/", basename(src))
     vsi_copy_file(src, src_mem)
 
